import numpy as np
# This is coded strictly according to the Rudary's description of the coversion
def Si(i, A, Q):
    if i == 0: Sn = np.zeros((len(A), len(A)))
    else:
        Sn = Q
        for k in range(i-1):
            Sn = Sn + np.linalg.matrix_power(A, k+1)*Q*np.linalg.matrix_power(np.transpose(A), k+1)
    return Sn
###########WHERE THE PROBLEM IS######################
def PhiBlock(i, j, A, Q, H):
    if j <= i: #j is corresponding to the row 
        block = H*np.linalg.matrix_power(A, i-j)*Si(j-1, A, Q)*np.transpose(H)
    else:
        return np.transpose(PhiBlock(j, i, A, Q, H))
        #block = H*np.linalg.matrix_power(A, j-i-1)*Si(i, A, Q)*np.transpose(H)
    return block
#####################################################

def PhiCol(j, n, A, Q, H):
    column = PhiBlock(1, j, A, Q, H)
    for i in range(2,n+1):
        column = np.concatenate((column, PhiBlock(i, j, A, Q, H)), axis=0)
    #print column
    return column

def Phi(n, A, Q, H):
    Matrix = PhiCol(1, n, A, Q, H)
    for j in range(2, n+1):
        Matrix = np.concatenate((Matrix, PhiCol(j, n, A, Q, H)), axis=1)
    return Matrix

def SigmaBlock(i, j, H, A, P, R, Q):
    if j>=i: # j is corresponding to column
        square = np.linalg.matrix_power(A, j-1)
        parta = H*square*P*np.transpose(H*np.linalg.matrix_power(A, i-1))
        partb = H*np.linalg.matrix_power(A, j-i)*Si(i-1, A, Q)*np.transpose(H)
        block = parta + partb
        if j == i:
            block = block + R
    else:
        block = SigmaBlock(j, i, H, A, P, R, Q)
    return block


def mu(n, H, A, x):
    mean = H*x
    for i in range(1, n):
        mean = np.concatenate((mean,H*np.linalg.matrix_power(A, i)*x), axis=0)
    return mean

def Sigma(n, H, A, P, R, Q):
    matrix = SigmaBlock(1, 1, H, A, P, R, Q)
    for j in range(2, n+1):
        matrix = np.concatenate((matrix, SigmaBlock(1, j, H, A, P, R, Q)), axis = 0)
    for i in range(2, n+1):
        col = SigmaBlock(i, 1, H, A, P, R, Q)
        for j in range(2, n+1):
            col = np.concatenate((col, SigmaBlock(i, j, H, A, P, R, Q)), axis = 0)
        matrix = np.concatenate((matrix, col), axis = 1)
    return matrix

def M(n, A, H):
    m = H
    for i in range(1, n):
        m = np.concatenate((m, H*np.linalg.matrix_power(A, i)), axis = 0)
    return m

# This code converts LDS to PLG, sigma is the noise covariance and Sigmaa is the initial mean covariance
def LDS2PSR(A, H, Q, R, x, P, diagR, n):
    Mu = mu(n, H, A, x)
    #print Mu
    #print M(n, A, H)
    G = H*np.linalg.matrix_power(A, n)*np.linalg.pinv(M(n, A, H))
    #print G
    Sigmaa = Sigma(n, H, A, P, R, Q)
    #print Sigmaa
    PhiC = PhiCol(n+1, n, A, Q, H)
    PHi = Phi(n,A, Q, H)
    #print PhiC
    #print PHi
    C = PhiC - PHi*np.transpose(G) - diagR*np.transpose(G)
    #print C
    sigma = H*Si(n, A, Q)*np.transpose(H)+R-G*PhiC - np.transpose(C)*np.transpose(G)
    #print sigma
    return (Mu, Sigmaa, G, C, sigma)
